import hashlib
import socket
import time
import logging

class TCPClientException(Exception):
    pass

class TCPClientStatuses:
    DISABLED = 0
    CONNECTED = 1
    RECONNECTING = 2

class TCPClient(object):

    def __init__(self, address: str, password: int, port: int=4711, reconnect: bool=True, debug: bool=True, debug_format:str="%(levelname)s:%(message)s", debug_level:object=logging.DEBUG):
        self.socket = None
        self.address = address
        self.password = password
        self.port = port
        self._reconnect = reconnect
        self.debug = debug
        self.debug_format = debug_format
        self.debug_level = debug_level
        self._status = TCPClientStatuses.DISABLED
        self._inr = 0

        logging.basicConfig(format=self.debug_format, level=self.debug_level)

    def log(self, *value):
        if self.debug: logging.debug(" ".join(str(v) for v in value))

    @property
    def status(self):
        return self._status

    @status.setter
    def status(self, status):
        self._status = status

    def start(self):
        self.socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.socket.connect((self.address, self.port))
        self.pre_init()
        self.status = TCPClientStatuses.CONNECTED
        return True

    def pre_init(self):
        welcResponse = ""
        while 1:
            data = self.socket.recv(1024)
            welcResponse += data.decode("utf-8")
            eosPos = welcResponse.find('\n\n')
            if eosPos != -1: break;

        prefix = '### Digest seed: '
        seedPos = welcResponse.find(prefix)
        seed = ""

        if seedPos == -1: raise Exception("Authorization is not possible")

        seedPos += len(prefix)
        seedPosEnd = welcResponse.find('\n', seedPos)
        seed = welcResponse[seedPos:seedPosEnd].encode("utf-8")

        md5Obj = hashlib.md5()
        md5Obj.update(seed)
        md5Obj.update(self.password.encode("utf-8"))
        passHash = md5Obj.hexdigest()

        self.socket.send(('\x02' + 'login ' + passHash + '\n').encode("utf-8"))
        data = self.socket.recv(1024).decode("utf-8")
        if "Authentication successful, rcon ready." not in data: raise Exception(data)

    def is_closed(self):
        return self.status == TCPClientStatuses.DISABLED

    def close(self):
        if not self.socket: return None
        self.status = TCPClientStatuses.DISABLED
        self.socket.close()

    def rcon_invoke(self, command):
        rt = time.time_ns()
        if not self.socket: raise Exception("The client is not connected")
        self.socket.send(('\x02' + command + '\n').encode("utf-8"))
        self._inr += 1
        result = ""
        done = False
        while not done:
            try:
                data = self.socket.recv(2048)
                if data is None: raise Exception("Client has terminated the current connection. ")
            except Exception as error:
                self.status = TCPClientStatuses.DISABLED
                raise Exception(error)
            for c in data:
                if c == 0x4:
                    done = True
                    break
                result += chr(c)
        if len(result) > 0 and result[-1] == "\n": result = result[:-1]
        ct = time.time_ns()
        #self.log(ct-rt, "ns")
        if result.strip() == "": return None
        return result


import re

class Map():
    def __init__(self, client, id, name, gpm, size):
        self.__client = client
        self.id = id
        self.name = name
        self.gpm = gpm
        self.size = size

    def __repr__(self):
        return "Map(name=%s, gpm=%s, size=%s)" % (self.name, self.gpm, self.size)

    def run(self):
        self.__client.rcon_invoke("exec admin.nextLevel %s" % self.id)
        result = self.__client.rcon_invoke("exec admin.runNextLevel")
        return result

    def set_next(self):
        result = self.__client.rcon_invoke("exec admin.nextLevel %s" % self.id)
        return result

    def remove(self):
        result = self.__client.rcon_invoke("exec mapList.remove %s" % self.id)
        return result

    def __str__(self):
        return self.name

    def __int__(self):
        return self.id

class MapList():
    def __init__(self, client):
        self.__client = client

    def __load(self):
        data = self.__client.rcon_invoke("exec maplist.list")
        list = {}
        pattern = re.compile('(\d+): "(\S*?)" (\S*) (\d+)')
        for map in data.split("\n"):
            matches = pattern.findall(map)
            if len(matches) != 0:
                mapID = int(matches[0][0])
                list[mapID] = Map(self.__client, mapID, matches[0][1], matches[0][2], matches[0][3])

        if len(list) < 1: return None
        return list

    @property
    def list(self):
        return self.__load()

    @property
    def total(self):
        return int(self.__client.rcon_invoke("exec mapList.mapCount"))

    @property
    def current(self):
        return self.list.get(int(self.__client.rcon_invoke("exec mapList.currentMap")))

    def clear(self):
        result = self.__client.rcon_invoke("exec mapList.clear")
        return result

    def remove(self, id: int):
        result = self.__client.rcon_invoke("exec mapList.remove %s" % id)
        return result

    def append(self, name: str, game_mode: str, size: int):
        result = self.__client.rcon_invoke("exec mapList.append \"%s\" %s %s" % (name, game_mode, size))
        return result

    def insert(self, id: int, name: str, game_mode: str, size: int):
        result = self.__client.rcon_invoke("exec mapList.insert %s \"%s\" %s %s" % (id, name, game_mode, size))
        return result

    def save(self):
        result = self.__client.rcon_invoke("exec mapList.save")
        return result


import re

class Ban():
    def __init__(self, client, type=None, address=None, period=None, key=None, **kwargs):
        self.__client = client
        self.type = type
        self.address = address
        self.period = period
        self.key = key
        for option, value in kwargs.items():
            setattr(self, option, value)

    def __repr__(self):
        if self.type == "address": return "Ban(address=%s, period=%s)" % (self.address, self.period)
        elif self.type == "key": return "Ban(key=%s, period=%s)" % (self.key, self.period)
        else: return "Ban(key=%s, address=%s, period=%s, type=%s)" % (self.key, self.address, self.period, self.type)

    def unban(self):
        if self.type == "address": return self.__client.rcon_invoke("exec admin.removeAddressFromBanList %s" % self.address)
        elif self.type == "key": return self.__client.rcon_invoke("exec admin.removeKeyFromBanList %s" % self.key)
        else: return False

class BanManager():
    def __init__(self, client):
        self.__client = client

    def __load(self):
        blAddrs = self.__client.rcon_invoke("exec admin.listBannedAddresses")
        blKeys = self.__client.rcon_invoke("exec admin.listBannedKeys")

        banlist = []

        pattern = re.compile('(\S*?): (\S*?) (\S*)')

        if blAddrs is not None:
            for ban in blAddrs.split("\n"):
                matches = pattern.findall(ban)
                if len(matches) != 0:
                    banlist.append(Ban(self.__client, type="address", address=matches[0][1], period=matches[0][2], key=None))

        if blKeys is not None:
            for ban in blKeys.split("\n"):
                matches = pattern.findall(ban)
                if len(matches) != 0:
                    banlist.append(Ban(self.__client, type="key", key=matches[0][1], period=matches[0][2], address=None))

        if len(banlist) < 1: return None
        return banlist

    @property
    def list(self):
        return self.__load()

    def clear(self):
        return self.__client.rcon_invoke("exec admin.clearBanList")

    def kick(self, id: int):
        return self.__client.rcon_invoke("exec admin.kickPlayer %s" % id)

    def add_ban(self, address: str, period="perm"):
        return self.__client.rcon_invoke("exec admin.addAddressToBanList %s %s" % (address, period))

    def add_ban_key(self, key: str, period="perm"):
        return self.__client.rcon_invoke("exec admin.addKeyToBanList %s %s" % (key, period))


from setuptools import setup, find_packages

def readme():
  with open('README.md', 'r') as f:
    return f.read()

setup(
  name='rconnet',
  version='0.0.1',
  author='VordyV',
  author_email='vordy.production@gmail.com',
  description='Python RCON client for the Battlefield 2142 server',
  long_description_content_type='text/markdown',
  long_description=readme(),
  url='https://github.com/VordyV/rconnet',
  packages=find_packages(),
  keywords='rcon client',
  python_requires='>=3.11'
)


import re

class Settings():
    def __init__(self, client):
        self.__client = client

    def __is_valid_ip(self, ip):
        ip_regex = r"^\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3}$"
        if re.match(ip_regex, ip):
            return True
        else:
            return False

    def __setting(self, type_option: str, option: str, value: object):

        if type_option == "str":
            if value is None:
                return self.__client.rcon_invoke("exec %s" % (option))
            return self.__client.rcon_invoke("exec %s \"%s\"" % (option, value))
        elif type_option == "bool":
            if value is None:
                return bool(int(self.__client.rcon_invoke("exec %s" % (option))))
            if bool(value) is True:
                value = "1"
            else:
                value = "0"
            return self.__client.rcon_invoke("exec %s \"%s\"" % (option, value))
        elif type_option == "int":
            if value is None:
                return self.__client.rcon_invoke("exec %s" % (option))
            if not isinstance(value, int): raise Exception("Wrong data type. Integer")
            return self.__client.rcon_invoke("exec %s \"%s\"" % (option, value))
        elif type_option == "float":
            if value is None:
                return self.__client.rcon_invoke("exec %s" % (option))
            if not isinstance(value, float): raise Exception("Wrong data type. Float")
            return self.__client.rcon_invoke("exec %s \"%s\"" % (option, value))

    def server_name(self, value: str=None):
        return self.__setting("int", "sv.serverName", value)

    def password(self, value: str=None):
        return self.__setting("str", "sv.password", value)

    def internet(self, value: bool=None):
        return self.__setting("bool", "sv.internet", value)

    def bandwidth_choke(self, value: bool=None):
        return self.__setting("bool", "sv.bandwidthChoke", value)

    def server_ip(self, value: str=None):
        if value is not None and not self.__is_valid_ip(value): Exception("IP address of the server is incorrect.")
        return self.__setting("str", "sv.serverIP", value)

    def server_port(self, value: int=None):
        return self.__setting("int", "sv.serverPort", value)

    def welcome_message(self, value: str=None):
        return self.__setting("str", "sv.welcomeMessage", value)

    def punk_buster(self, value: bool=None):
        return self.__setting("bool", "sv.punkBuster", value)

    def allow_free_cam(self, value: bool=None):
        return self.__setting("bool", "sv.allowFreeCam", value)

    def allow_external_views(self, value: bool=None):
        return self.__setting("bool", "sv.allowExternalViews", value)

    def allow_nose_cam(self, value: bool=None):
        return self.__setting("bool", "sv.allowNoseCam", value)

    def hit_indicator(self, value: bool=None):
        return self.__setting("bool", "sv.hitIndicator", value)

    def max_players(self, value: int=None):
        return self.__setting("int", "sv.maxPlayers", value)

    def num_players_needed_to_start(self, value: int=None):
        return self.__setting("int", "sv.numPlayersNeededToStart", value)

    def not_enough_players_restart_delay(self, value: int=None):
        return self.__setting("int", "sv.notEnoughPlayersRestartDelay", value)

    def start_delay(self, value: int=None):
        return self.__setting("int", "sv.startDelay", value)

    def end_delay(self, value: int=None):
        return self.__setting("int", "sv.endDelay", value)

    def spawn_time(self, value: int=None):
        return self.__setting("int", "sv.spawnTime", value)

    def man_down_time(self, value: int=None):
        return self.__setting("int", "sv.manDownTime", value)

    def end_round_delay(self, value: int=None):
        return self.__setting("int", "sv.endOfRoundDelay", value)

    def ticket_ratio(self, value: int=None):
        return self.__setting("int", "sv.ticketRatio", value)

    def rounds_per_map(self, value: int=None):
        return self.__setting("int", "sv.roundsPerMap", value)

    def time_limit(self, value: int=None):
        return self.__setting("int", "sv.timeLimit", value)

    def score_limit(self, value: int=None):
        return self.__setting("int", "sv.scoreLimit", value)

    def soldier_friendly_fire(self, value: int=None):
        return self.__setting("int", "sv.soldierFriendlyFire", value)

    def vehicle_friendly_Fire(self, value: int=None):
        return self.__setting("int", "sv.vehicle_friendly_Fire", value)

    def soldier_splash_friendly_fire(self, value: int=None):
        return self.__setting("int", "sv.soldierSplashFriendlyFire", value)

    def vehicle_splash_friendly_fire(self, value: int=None):
        return self.__setting("int", "sv.vehicleSplashFriendlyFire", value)

    def tk_punish_enabled(self, value: bool=None):
        return self.__setting("bool", "sv.tkPunishEnabled", value)

    def tk_num_punish_to_kick(self, value: int=None):
        return self.__setting("int", "sv.tkNumPunishToKick", value)

    def tk_punish_by_default(self, value: int=None):
        return self.__setting("int", "sv.tkPunishByDefault", value)

    def voting_enabled(self, value: bool=None):
        return self.__setting("bool", "sv.votingEnabled", value)

    def vote_time(self, value: int=None):
        return self.__setting("int", "sv.voteTime", value)

    def min_players_for_voting(self, value: int=None):
        return self.__setting("int", "sv.minPlayersForVoting", value)

    def game_spy_port(self, value: int=None):
        return self.__setting("int", "sv.gameSpyPort", value)

    def allow_NAT_Negotiation(self, value: bool=None):
        return self.__setting("bool", "sv.allowNATNegotiation", value)

    def interface_IP(self, value: str=None):
        if value is not None and not self.__is_valid_ip(value): Exception("Interface IP is incorrect.")
        return self.__setting("str", "sv.interfaceIP", value)

    def auto_record(self, value: bool=None):
        return self.__setting("bool", "sv.autoRecord", value)

    def demo_index_URL(self, value: str=None):
        return self.__setting("str", "sv.demoIndexURL", value)

    def demo_download_URL(self, value: str=None):
        return self.__setting("str", "sv.demoDownloadURL", value)

    def auto_demo_hook(self, value: str=None):
        return self.__setting("str", "sv.autoDemoHook", value)

    def demo_Quality(self, value: int=None):
        return self.__setting("int", "sv.demoQuality", value)

    def admin_script(self, value: str=None):
        return self.__setting("str", "sv.adminScript", value)

    def time_before_restart_map(self, value: int=None):
        return self.__setting("int", "sv.timeBeforeRestartMap", value)

    def auto_balance_team(self, value: bool=None):
        return self.__setting("bool", "sv.autoBalanceTeam", value)

    def team_ratio_percent(self, value: int=None):
        return self.__setting("int", "sv.teamRatioPercent", value)

    def voip_enabled(self, value: bool=None):
        return self.__setting("bool", "sv.voipEnabled", value)

    def voip_quality(self, value: int=None):
        return self.__setting("int", "sv.voipQuality", value)

    def voip_server_remote(self, value: bool=None):
        return self.__setting("bool", "sv.voipServerRemote", value)

    def voip_server_remote_IP(self, value: str=None):
        if value is not None and not self.__is_valid_ip(value): Exception("Voip server remote IP is incorrect.")
        return self.__setting("str", "sv.voipServerRemoteIP", value)

    def voip_server_port(self, value: int=None):
        return self.__setting("int", "sv.voipServerPort", value)

    def voip_BF_client_port(self, value: int=None):
        return self.__setting("int", "sv.voipBFClientPort", value)

    def voip_BF_server_Port(self, value: int=None):
        return self.__setting("int", "sv.voipBFServerPort", value)

    def voip_hared_password(self, value: str=None):
        return self.__setting("str", "sv.voipSharedPassword", value)

    def use_global_rank(self, value: bool=None):
        return self.__setting("bool", "sv.useGlobalRank", value)

    def use_global_unlocks(self, value: bool=None):
        return self.__setting("bool", "sv.useGlobalUnlocks", value)

    def min_unlock_level(self, value: int=None):
        return self.__setting("int", "sv.minUnlockLevel", value)

    def max_unlock_level(self, value: int=None):
        return self.__setting("int", "sv.maxUnlockLevel", value)

    def sponsor_text(self, value: str=None):
        return self.__setting("str", "sv.sponsorText", value)

    def sponsor_logo_url(self, value: str=None):
        return self.__setting("str", "sv.sponsorLogoURL", value)

    def community_logo_url(self, value: str=None):
        return self.__setting("str", "sv.communityLogoURL", value)

    def custom_map_url(self, value: int=None):
        return self.__setting("int", "sv.customMapURL", value)

    def radio_spam_interval(self, value: int=None):
        return self.__setting("int", "sv.radioSpamInterval", value)

    def radio_max_spam_flag_count(self, value: int=None):
        return self.__setting("int", "sv.radioMaxSpamFlagCount", value)

    def radio_blocked_duration_time(self, value: int=None):
        return self.__setting("int", "sv.radioBlockedDurationTime", value)

    def max_rank(self, value: int=None):
        return self.__setting("int", "sv.maxRank", value)

    def allow_spectators(self, value: bool=None):
        return self.__setting("bool", "sv.allowSpectators", value)

    def allow_titan_movement(self, value: int=None):
        return self.__setting("int", "sv.allowTitanMovement", value)

    def bot_skill(self, value: float=None):
        return self.__setting("float", "sv.botSkill", value)


from rconnet.tcpclient import TCPClient
from rconnet.settings import Settings
from rconnet.maplist import MapList
from rconnet.banmanager import BanManager
import re

class Player():
    def __init__(self, client, id, nick, key=None, address=None, port=None, **kwargs):
        self._client = client
        self.id = id
        self.nick = nick
        self.key = key
        self.address = address
        self.port = port
        for option, value in kwargs.items():
            setattr(self, option, value)

    def __int__(self):
        return self.id

    def __str__(self):
        return self.nick

    def kick(self):
        return self._client.rcon_invoke("exec admin.kickPlayer %s" % self.id)

    def ban(self, period="perm"):
        return self._client.rcon_invoke("exec admin.banPlayer %s %s" % (self.id, period))

    def ban_key(self, period="perm"):
        if self.key is None: raise Exception("The player %s does not have a key" % (self.id))
        return self._client.rcon_invoke("exec admin.banPlayerKey %s %s" % (self.id, period))

class Default(TCPClient):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.settings = Settings(self)
        self.maplist = MapList(self)
        self.banmanager = BanManager(self)

    def __enter__(self):
        self.start()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if not self.is_closed():
            self.close()

    @property
    def users(self):
        return [user.replace(" ", "").split(":") for user in self.rcon_invoke("users").split("\n")[1:]]

    def __get_players(self):
        rawData = self.rcon_invoke('exec admin.listplayers')
        pattern1 = re.compile(r'''^Id:\ +(\d+)\ -\ (.*?)\ is\ remote\ ip:\ (\d+\.\d+\.\d+\.\d+):(\d+)''', re.VERBOSE)
        pattern2 = re.compile(r'''(?:.*hash:\ (\w{32}))?''', re.VERBOSE)
        players = {}

        i = 0
        for line in rawData.split("\n"):
            if i == 0:
                matches = pattern1.findall(line)
                if len(matches) != 0:
                    p_id = int(matches[0][0])

                    players[p_id] = Player(self, p_id, matches[0][1], address=matches[0][2], port=matches[0][3])

            elif i == 1:
                matches = pattern2.findall(line)
                players[p_id].key = matches[0]

            i ^= 1

        if len(players) < 1: return None

        return players

    @property
    def players(self):
        return self.__get_players()

    def run_next_level(self):
        return self.rcon_invoke("exec admin.runNextLevel")

    def run_level(self, name: str, game_mode: str, size: int):
        return self.rcon_invoke("exec admin.runLevel %s %s %s" % (name, game_mode, size))

    def current_level(self):
        return self.rcon_invoke("exec admin.currentLevel")

    def set_next_level(self, id: int):
        return self.rcon_invoke("exec admin.nextLevel %s" % id)

    def restart_map(self):
        return self.rcon_invoke("exec admin.restartMap")

    def say(self, message):
        return self.rcon_invoke("exec game.sayall \"%s\"" % message)